import random
import PileFile as pf
from graphviz import Digraph

class Arbre:
    def __init__(self, info=None, fg=None, fd=None):
        self.info = info
        self.fg = fg
        self.fd = fd

    def estVide(self):
        return self.info == None

    def insererElement(self, e):
        if self.estVide():
            self.info = e
        else :
            if e < self.info :
                if self.fg :
                    self.fg.insererElement(e)
                else:
                    self.fg = Arbre(e)
            if e > self.info :
                if self.fd :
                    self.fd.insererElement(e)
                else:
                    self.fd = Arbre(e)

    def afficherParcoursInfixe(self):
        if self.estVide() : return
        if self.fg : self.fg.afficherParcoursInfixe()
        print(self.info,end=' ')
        if self.fd : self.fd.afficherParcoursInfixe()

    def afficherParcoursPrefixe(self):
        if self.estVide() : return
        print(self.info,end=' ')
        if self.fg : self.fg.afficherParcoursPrefixe()
        if self.fd : self.fd.afficherParcoursPrefixe()

    def afficherParcoursPostfixe(self):
        if self.estVide() : return
        if self.fg : self.fg.afficherParcoursPostfixe()
        if self.fd : self.fd.afficherParcoursPostfixe()
        print(self.info,end=' ')

    def afficherArbre(self):
        print()
        lignes, *_ = self.afficherArbreRec()
        for ligne in lignes:
            print(ligne)

    def afficherArbreRec(self):
        """Retourne une ligne de strings, largueur, hauteur, et coordonnee horizontale de la racine"""
        # Le noeud n'a pas de fils
        if self.fd == None and self.fg == None:
            ligne = '%s' % self.info
            largueur = len(ligne)
            hauteur = 1
            milieu = largueur // 2
            return [ligne], largueur, hauteur, milieu

        # Fils gauche uniquement
        if self.fd == None:
            lignes, l, h, c = self.fg.afficherArbreRec()
            s = '%s' % self.info
            u = len(s)
            premiere_ligne = (c + 1) * ' ' + (l - c - 1) * '_' + s
            seconde_ligne = c * ' ' + '/' + (l - c - 1 + u) * ' '
            lignes_decalees = [ligne + u * ' ' for ligne in lignes]
            return [premiere_ligne, seconde_ligne] + lignes_decalees, l + u, h + 2, l + u // 2

        # Fils droit uniquement
        if self.fg == None:
            lignes, l, h, c = self.fd.afficherArbreRec()
            s = '%s' % self.info
            u = len(s)
            premiere_ligne = s + c * '_' + (l - c) * ' '
            seconde_ligne = (u + c) * ' ' + '\\' + (l - c - 1) * ' '
            lignes_decalees = [u * ' ' + ligne for ligne in lignes]
            return [premiere_ligne, seconde_ligne] + lignes_decalees, l + u, h + 2, u // 2


        # Deux fils
        gauche, n, p, x = self.fg.afficherArbreRec()
        droit, m, q, y = self.fd.afficherArbreRec()
        s = '%s' % self.info
        u = len(s)
        premiere_ligne = (x + 1) * ' ' + (n - x - 1) * '_' + s + y * '_' + (m - y) * ' '
        seconde_ligne = x * ' ' + '/' + (n - x - 1 + u + y) * ' ' + '\\' + (m - y - 1) * ' '
        if p < q:
            gauche += [n * ' '] * (q - p)
        elif q < p:
            droit += [m * ' '] * (p - q)
        lignes_zippees = zip(gauche, droit)
        lignes = [premiere_ligne, seconde_ligne] + [a + u * ' ' + b for a, b in lignes_zippees]
        return lignes, n + m + u, max(p, q) + 2, n + u // 2


    def affichage_pdf(self, nom="arbre.pdf"):
        if self.estVide(): return
        res = Digraph()
        res.node(str(self.info),str(self.info))
        self.affichage_pdf_rec(res)
        res.render(nom,view=True)

    def affichage_pdf_rec(self,res):
        if self.fg:
            res.node(str(self.fg.info),str(self.fg.info))
            res.edge(str(self.info),str(self.fg.info))
            self.fg.affichage_pdf_rec(res)
        else : # pour ameliorer l'affichage (ajoute un espace vide dans le fils qui n'existe pas)
            res.node(str(self.info)+'_fg',str(self.info)+'_fg',color='white',fontcolor='white')
            res.edge(str(self.info),str(self.info)+'_fg',color='white',fontcolor='white')

        if self.fd:
            res.node(str(self.fd.info),str(self.fd.info))
            res.edge(str(self.info),str(self.fd.info))
            self.fd.affichage_pdf_rec(res)
        else : # pour ameliorer l'affichage (ajoute un espace vide dans le fils qui n'existe pas)
            res.node(str(self.info)+'_fd',str(self.info)+'_fd',color='white',fontcolor='white')
            res.edge(str(self.info),str(self.info)+'_fd',color='white',fontcolor='white')


    def rechercherElement(self, e):
        if self.estVide(): return (False, None)
        if e == self.info: return (True, self)
        if e < self.info:
            if self.fg:
                return self.fg.rechercherElement(e)
            else:
                return (False, None)
        if e > self.info:
            if self.fd:
                return self.fd.rechercherElement(e)
            else:
                return (False, None)

    def rechercherElementIteratif(self, e):
        if self.estVide(): return (False, None)
        f = pf.File([])
        f.enfiler(self)
        while not f.estVide():
            n = f.traiter()
            if n.info == e: return (True, n)
            if n.fg: f.enfiler(n.fg)
            if n.fd: f.enfiler(n.fd)
        return (False, None)

    def hauteurArbre(self):
        if self.estVide(): return -1
        hgauche = self.fg.hauteurArbre() if self.fg else -1
        hdroit = self.fd.hauteurArbre() if self.fd else -1
        if hgauche > hdroit:
            return hgauche + 1
        else:
            return hdroit + 1

    def afficherParcoursLargeur(self):
        if self.estVide(): return
        f = pf.File([])
        f.enfiler(self)
        while not f.estVide():
            n = f.traiter()
            if n.fg: f.enfiler(n.fg)
            if n.fd: f.enfiler(n.fd)
            print(n.info, end=' ')

    def afficherParcoursPrefixeIteratif(self):
        if self.estVide(): return
        p = pf.Pile([])
        p.empiler(self)
        while not p.estVide():
            n = p.sommet()
            p.depiler()
            print(n.info, end=' ')
            if n.fd: p.empiler(n.fd)
            if n.fg: p.empiler(n.fg)

    def estEgal(self, arbre):
        if self.estVide() or arbre.estVide(): return self.estVide() and arbre.estVide()  # au moins 1 vide
        if self.info != arbre.info:  # valeurs differentes
            return False
        fgOk = True
        if self.fg:
            if arbre.fg:
                fgOk = self.fg.estEgal(arbre.fg)  # les deux fg existent, on continue
            else:
                return False  # fg self exite mais pas fg arbre
        else:
            fgOk = (arbre.fg == None)  # fg self n'existe pas, il faut que fg arbre n'existe pas
        fdOk = True
        if self.fd:
            if arbre.fd:
                fdOk = self.fd.estEgal(arbre.fd)  # les deux fd existent, on continue
            else:
                return False  # fd self exite mais pas fd arbre
        else:
            fdOk = (arbre.fd == None)  # fd self n'existe pas, il faut que fg arbre n'existe pas
        return fgOk and fdOk  # les deux fils doivent etre egaux


    def supprimerElement(self, e):
        if self.estVide(): return  # rien a faire si arbre vide
        n = self
        pere = None
        trouve = False
        explorer_gauche = True
        # on commence par cherche l'element (et son pere et quel fils il est)
        while n and not trouve:
            if n.info == e:
                trouve = True
            else:
                pere = n
                if e < n.info:
                    explorer_gauche = True
                    n = n.fg
                else:
                    explorer_gauche = False
                    n = n.fd
        if not trouve: return  # cas 0, pas d'effet si pas present
        # l'element est present
        if not n.fg and not n.fd:
            # cas a : c'est une feuille
            if not pere:
                # on supprime la racine
                n.info = None
            else:
                # sinon on vient de la gauche ou de la droite
                if explorer_gauche:
                    pere.fg = None
                else:
                    pere.fd = None
        else:
            # cas b1 : n a seulement un fils gauche
            if n.fg and not n.fd:
                if not pere:
                    # on supprime la racine
                    n.info = n.fg.info
                    n.fg = n.fg.fg
                    n.fd = n.fg.fd
                else:
                    # on court-circuite le noeud
                    if explorer_gauche:
                        pere.fg = n.fg
                    else:
                        pere.fd = n.fg
            else:
                # cas b2 : n a seulement un fils droit
                if not n.fg and n.fd:
                    if not pere:
                        # on supprime la racine
                        n.info = n.fd.info
                        n.fg = n.fd.fg
                        n.fd = n.fd.fd
                    else:
                        # on court-circuite le noeud
                        if explorer_gauche:
                            pere.fg = n.fd
                        else:
                            pere.fd = n.fd
                else:
                    # cas c : n a deux fils
                    # on cherche le plus proche successeur
                    # cad le noeud le plus a gauche du sous arbre droit
                    successeur = n.fd
                    pere_successeur = n
                    while successeur.fg:
                        pere_successeur = successeur
                        successeur = successeur.fg
                    n.info = successeur.info  # pas besoin de changer fg et fd, juste les infos
                    if pere_successeur == n:
                        # le sous-arbre droit etant réduit à une seule feuille, on n'est pas rentré
                        # dans le while, il faut changer le fd de n
                        pere_successeur.fd = successeur.fd
                    else:
                        # on est passé au moins une fois dans le while, c'est le fg d'un des noeuds
                        # du sous-arbre droit qu'il faut changer
                        pere_successeur.fg = successeur.fd  # on court-circuite le successeur



    def estABR(self):
        if self.estVide(): return True
        (test, min, max) = self.__estABRSousArbre()
        return test

    def __estABRSousArbre(self):
        minG = maxG = minD = maxD = self.info
        testG = testD = True
        if self.fg: (testG, minG, maxG) = self.fg.__estABRSousArbre()
        if self.fd: (testD, minD, maxD) = self.fd.__estABRSousArbre()
        if not testG or not testD: return (False, None, None)
        if maxG > self.info or minD < self.info: return (False, None, None)
        return (True, minG, maxD)

    def maxBrancheGlouton(self):
        if self.estVide(): return None
        if self.fg and self.fd :
            # noeud a deux fils
            if self.fg.info > self.fd.info:
                return self.info + self.fg.maxBrancheGlouton()
            else:
                return self.info + self.fd.maxBrancheGlouton()
        else:
            if self.fg == None and self.fd == None:
                # feuille
                return self.info
            else :
                if self.fg:
                    # fg existe, fd non
                    return self.info + self.fg.maxBrancheGlouton()
                if self.fd:
                    # fd existe, fg non
                    return self.info + self.fd.maxBrancheGlouton()

    def maxBrancheProgDyn(self):
        if self.estVide(): return None
        sum_fg = self.info if self.fg == None else self.info+self.fg.maxBrancheProgDyn()
        sum_fd = self.info if self.fd == None else self.info+self.fd.maxBrancheProgDyn()
        return max(sum_fg, sum_fd)

# ================= programme principal ======================
arbre = Arbre()

print('Arbre vide ? ',arbre.estVide())

print("Ajouts des elements 5 2 4 6 : ", end='')
arbre.insererElement(5)
arbre.insererElement(2)
arbre.insererElement(4)
arbre.insererElement(6)

arbre.afficherParcoursInfixe()
print()

arbre.afficherArbre()

#arbre.affichage_pdf()

print("Ajout d'elements aleatoires : ", end='')
for _ in range(20):
    arbre.insererElement(int(random.random() * 20) - 10)

print()
arbre.afficherArbre()
arbre.afficherParcoursInfixe()
print()
arbre.afficherParcoursPrefixe()
print()
arbre.afficherParcoursPostfixe()
print()
#arbre.affichage_pdf()

print("Recherche (rec) de 4 : ", end='')
(trouve, noeud) = arbre.rechercherElement(4)
if trouve:
    print("trouve : id=" + str(id(noeud)))
else:
    print("non trouve")
print("Recherche (rec) de 11 : ", end='')
(trouve, noeud) = arbre.rechercherElement(11)
if trouve:
    print("trouve : id=" + str(id(noeud)))
else:
    print("non trouve")

print("Recherche (ite) de 4 : ", end='')
(trouve, noeud) = arbre.rechercherElementIteratif(4)
if trouve:
    print("trouve : id=" + str(id(noeud)))
else:
    print("non trouve")
print("Recherche (ite) de 11 : ", end='')
(trouve, noeud) = arbre.rechercherElementIteratif(11)
if trouve:
    print("trouve : id=" + str(id(noeud)))
else:
    print("non trouve")

print("Hauteur de l'arbre : ", end='')
print(str(arbre.hauteurArbre()))

print("Affichage en largeur : ", end='')
arbre.afficherParcoursLargeur()
print()

print("Affichage prefixe iteratif : ", end='')
arbre.afficherParcoursPrefixeIteratif()
print()

arbre1 = Arbre()
arbre2 = Arbre()
for _ in range(20):
    nb = int(random.random() * 20) - 10
    arbre1.insererElement(nb)
    arbre2.insererElement(nb)
if arbre1.estEgal(arbre2):
    print("Les deux arbres sont egaux")
else:
    print("Les deux arbres sont differents")
arbre2.insererElement(11)
if arbre1.estEgal(arbre2):
    print("Les deux arbres sont egaux")
else:
    print("Les deux arbres sont differents")

if arbre.estABR():
    print("L'arbre est un ABR")
else:
    print("L'arbre n'est pa s un ABR")

arbre3 = Arbre()
arbre3.insererElement(10)
arbre3.insererElement(5)
arbre3.insererElement(15)
arbre3.insererElement(3)
arbre3.insererElement(8)
arbre3.insererElement(20)
arbre3.insererElement(1)
arbre3.insererElement(4)
arbre3.insererElement(9)
arbre3.insererElement(16)
print("Arbre : ")
arbre3.afficherArbre()
print()
print("Suppression de 5 : ")
arbre3.supprimerElement(5)
arbre3.afficherArbre()
print()
print("Suppression de 10 : ")
arbre3.supprimerElement(10)
arbre3.afficherArbre()
print()
print("Ajout des elements 5 10 : ")
arbre3.insererElement(5)
arbre3.insererElement(10)
arbre3.afficherArbre()
print()

print("Somme max des branches (glouton) : "+str(arbre3.maxBrancheGlouton()))
print("Suppression de 16 et ajout de 12 : ")
arbre3.supprimerElement(16)
arbre3.insererElement(12)
arbre3.afficherArbre()
print()
print("Somme max des branches (glouton) : "+str(arbre3.maxBrancheGlouton()))
print("Somme max des branches (prog dyn) : "+str(arbre3.maxBrancheProgDyn()))